/* DMA.H - Programmed by Dennis Ranke and Jonathan Hobson */
/* 250704: Maarten Hofman: Fiddled with the DMA */
/* 270904: Maarten Hofman: Added redo from start for static chains */

#ifndef __DMA_H__
#define __DMA_H__

#include "sps2wrap.h"
#include "singleton.h"
#include "ps2matrix4x4.h"
#include "ps2vector4.h"
#include "ps2defines.h"

#define VIFStaticDMA CVIFStaticDMA::GetSingleton()
#define VIFDynamicDMA CVIFDynamicDMA::GetSingleton()

#define GIFStaticDMA CGIFStaticDMA::GetSingleton()
#define GIFDynamicDMA CGIFDynamicDMA::GetSingleton()

class CDMAChain
{
	public:
		void Begin();				// Starts a new DMA chain

		void Add32(uint32 data)
		{
			if(m_pPtr >= m_pEndPtr) {
                          /* We will be writing to the new page, so stitch */
                          Stitch();
                        }  
			*m_pPtr++ = data;		// Write the data and move the pointer on
		}

		void Add64(uint64 data)
		{ Add32((uint32)data);	Add32((uint32)(data >> 32)); }

		void Add128(uint128 data)
		{ Add64((uint64)data);	Add64((uint64)(data >> 64)); }

		void AddFloat(float data)
		{ Add32(*(uint32 *)&data); }

		void AddVectorI(int x, int y, int z, int w);
		void AddVectorF(float x, float y, float z, float w);

		void AddVector(const KazMaths::Vector4 & Vector)
		{
			AddFloat(Vector.x);
			AddFloat(Vector.y);
			AddFloat(Vector.z);
			AddFloat(Vector.w);
		}

		void AddMatrix(const KazMaths::Matrix4x4 & Matrix);

		void Align(int iAlign, int iOffset = 0);	// Aligns the write pointer to a mulitple of iAlign * 32 bits
									// The value can be offset by iOffset * 32 bits

		void PrepForDMATag();		// Makes sure that we are writing a multiple of 1 quadword and also makes sure
									// that the next tag isn't on a stitching boundary

		// DMA Tags
		void DMAEnd();
		void DMARet();
		void DMACall(uint32 iAddr);

		void StartDirect();
		void EndDirect();

		void StartAD();
		void EndAD(){};
		void AddAD(uint64 iData, uint64 iAddr);

		void StartMPG(uint32 iAddr);
		void EndMPG();
		void AddMPG(uint64 iInst);

		void AddUnpack(int format, int addr, int num, int usetops = 0, int nosign = 1, int masking = 0)
		{
			Add32((0x60 << 24) + (format << 24) + (masking << 28) + (usetops << 15) +
				(nosign << 14) + (num << 16) + addr);
		}

		virtual void PrintPacket();

		int	m_iPath3SyncCodes;
		
	protected:
		virtual int NewPage() = 0;	// Starts writing to a new memory page
		virtual void Stitch();		// Stitches across the page boundary with a next tag
		virtual void NewTag();		// Reserves space in the chain for a DMA tag
		
		int m_iNumPages;			// How many pages we have allocated for DMAing
		CDMAMem * m_pPages;			// All the pages that we have allocated

		int * m_pPtr;				// The data write pointer
		int * m_pEndPtr;			// The next page starts here
		int m_iPhysAddr;			// The physical address of the first page (for TADR)

		uint64 * m_pDMATag;			// The currently open DMA tag
		int * m_pVIFCode;			// The currently open VIF code

		int m_iMPGCount;			// How many MPG instructions we have written
		int m_iMPGAddr;				// The VU_MICRO_MEM address we are writing to

		uint64 * m_pADGiftag;
		
		typedef enum
		{
			VIF_BASE,
			VIF_DIRECT				// Direct + TTE has alignment issues, MPG, and unpack do not
		} VIF1_STATE;				// The different states the DMA chain can be in

		VIF1_STATE m_VifState;
};

class CDMAChainNoTTE : public CDMAChain
{
	protected:
		virtual void NewTag();
};

class CVIFStaticDMA : public CDMAChain, public CSingleton<CVIFStaticDMA>
{
	public:
		CVIFStaticDMA();
		virtual ~CVIFStaticDMA();

		void Initialise(int iNumPages);
                void redo_from_start();

		int GetPointer();			// Retrieve the current pointer's physical address

	protected:
		virtual int NewPage();

		int m_iCurrPage;
		int m_iPhysicalOffset;
};

inline int CVIFStaticDMA::GetPointer()
{
	return (int)m_pDMATag + m_iPhysicalOffset;
}

class CVIFDynamicDMA : public CDMAChain, public CSingleton<CVIFDynamicDMA>
{
	public:
		CVIFDynamicDMA();
		virtual ~CVIFDynamicDMA();

		void Initialise(int iNumPages);
		void Fire(bool bFire = true, bool bPrint = false);

		virtual void PrintPacket();

		void OpenPath3();

	protected:
		virtual int NewPage();
		
		int m_iCurrPage;
		int m_iCurrBuffer;
};

class CGIFStaticDMA : public CDMAChainNoTTE, public CSingleton<CGIFStaticDMA>
{
	public:
		CGIFStaticDMA();
		virtual ~CGIFStaticDMA();

		void Initialise(int iNumPages);
                void redo_from_start();

		int GetPointer();

	protected:
		virtual int NewPage();

		int m_iCurrPage;
		int m_iPhysicalOffset;
};

inline int CGIFStaticDMA::GetPointer()
{
	return (int)m_pDMATag + m_iPhysicalOffset;
}

class CGIFDynamicDMA : public CDMAChainNoTTE, public CSingleton<CGIFDynamicDMA>
{
	public:
		CGIFDynamicDMA();
		virtual ~CGIFDynamicDMA();

		void Initialise(int iNumPages);
		void Fire(bool bFire = true, bool bPrint = false);

		virtual void PrintPacket();

	protected:
		virtual int NewPage();
		
		int m_iCurrPage;
		int m_iCurrBuffer;
};

#endif
